/*
 *	Find Duplicate Files
 *
 *	(c) 1997 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "types.h"
#include "fastbuf.h"
#include "filemap.h"

static struct fastbuf *f;

struct dinfo {
  struct dinfo *parent;
  byte name[1];
};

struct finfo {
  struct dinfo *parent;
  struct finfo *next;
  ulg len, cs;
  byte name[1];
};

static struct finfo *first_file;

static void
scan(void)
{
  struct dinfo *root = NULL;
  struct dinfo *d;
  struct finfo *i, **lf = &first_file;
  int l, type;
  ulg len, chs;
  byte buf[MAX_NAME_SIZE];

  bread(f, buf, 4);
  if (memcmp(buf, FM_IDENT, 4) || bgetl(f) != FM_VERSION)
    die("%s: Invalid format", f->name);
  while ((type = bgetc(f)) >= 0)
    {
      if (type == E_EOD)
	{
	  if (root)
	    root = root->parent;
	  else
	    die("%s: Hollow tree", f->name);
	  continue;
	}
      l = bgetc(f);
      bread(f, buf, l);
      buf[l] =  0;
      bgetw(f);
      bgetw(f);
      bgetw(f);
      bgetl(f);
      switch (type)
	{
	case E_FILE:
 	  len = bgetl(f);
	  chs = 0;
	file:
	  i = xmalloc(sizeof(struct finfo) + strlen(buf));
	  strcpy(i->name, buf);
	  i->parent = root;
	  i->len = len;
	  i->cs = chs;
	  *lf = i;
	  lf = &i->next;
	  break;
	case E_FILE_CRC:
 	  len = bgetl(f);
 	  chs = bgetl(f);
	  goto file;
	case E_SYMLINK:
	case E_HARDLINK:
	  l = bgetc(f);
	  bseek(f, l, SEEK_CUR);
	  break;
	case E_CHAR:
	case E_BLOCK:
	  bgetl(f);
	  break;
	case E_DIR:
	  d = xmalloc(sizeof(struct dinfo) + strlen(buf));
	  strcpy(d->name, buf);
	  d->parent = root;
	  root = d;
	  break;
	case E_FIFO:
	case E_SOCK:
	  break;
	default:
	  die("%s: Invalid type %02x at %08x", f->name, type, btell(f)-1);
	}
    }
  if (root)
    die("%s: Malformed file");
  *lf = NULL;
}

static struct finfo *
sort(struct finfo *a, int crit)
{
  struct finfo *x, *y;
  struct finfo **xx, **yy;

  if (!a || !a->next)
    return a;
  xx = &x;
  yy = &y;
  while (a)
    {
      *xx = a;
      xx = &a->next;
      a = a->next;
      if (a)
	{
	  *yy = a;
	  yy = &a->next;
	  a = a->next;
	}
    }
  *xx = *yy = NULL;

  x = sort(x, crit);
  y = sort(y, crit);
  xx = &a;
  while (x && y)
    {
      if (crit ? (x->len < y->len) : (x->cs < y->cs))
	{
	  *xx = x;
	  xx = &x->next;
	  x = x->next;
	}
      else
	{
	  *xx = y;
	  xx = &y->next;
	  y = y->next;
	}
    }
  *xx = x ? x : y;
  return a;
}

static byte *
build_dir_name(struct dinfo *d, byte *b)
{
  if (d->parent)
    b = build_dir_name(d->parent, b);
  b += sprintf(b, "%s/", d->name);
  return b;
}

static void
build_name(struct finfo *f, byte *b)
{
  b = build_dir_name(f->parent, b);
  strcpy(b, f->name);
}

static void
doo(struct finfo *a)
{
  struct finfo *sg, *zg, *z, *la = NULL;
  byte na[MAX_NAME_SIZE], nb[MAX_NAME_SIZE];
  int f1, f2, r;
  void *b1 = NULL, *b2 = NULL;

  while (a)
    {
      sg = a;
      while (a && a->len == sg->len)
	{
	  la = a;
	  a = a->next;
	}
      la->next = NULL;
      sg = sort(sg, 0);
      while (sg)
	{
	  zg = sg;
	  while (sg && zg->cs == sg->cs)
	    sg = sg->next;
	  while (zg != sg && zg->next != sg)
	    {
	      if (zg->len != ~0)
		{
		  build_name(zg, na);
		  if (zg->len)
		    {
		      f1 = open(na, O_RDONLY);
		      if (f1 < 0)
			{
			  perror(na);
			  goto xx;
			}
		      b1 = mmap(NULL, zg->len, PROT_READ, MAP_PRIVATE, f1, 0);
		      if (b1 == (void *) -1)
			die("mmap failed: %m");
		    }
		  else
		    f1 = 0;
		  for(z=zg->next; z != sg; z=z->next)
		    {
		      if (z->len == ~0)
			continue;
		      build_name(z, nb);
		      if (f1)
			{
			  f2 = open(nb, O_RDONLY);
			  if (f2 < 0)
			    {
			      perror(nb);
			      continue;
			    }
			  b2 = mmap(NULL, z->len, PROT_READ, MAP_PRIVATE, f2, 0);
			  if (b2 == (void *) -1)
			    die("mmap failed: %m");
			  r = memcmp(b1, b2, z->len);
			  munmap(b2, z->len);
			  close(f2);
			  if (r)
			    continue;
			}
		      zg->len = z->len = ~0;
		      printf("%s == %s\n", na, nb);
		    }
		  if (f1)
		    {
		      munmap(b1, zg->len);
		      close(f1);
		    }
		xx:
		}
	      zg = zg->next;
	    }
	}
    }
}

int
main(int argc, char **argv)
{
  if (argc != 2)
    {
      fprintf(stderr, "Usage: fdups <map>\n");
      return 1;
    }
  f = bopen(argv[1], O_RDONLY, 65536);
  scan();
  first_file = sort(first_file, 1);
  doo(first_file);
  return 0;
}
